@@ -3,59 +3,25 @@
 
 XBase::Memo - Generic support for various memo formats
 
-=head1 SYNOPSIS
-
-Used indirectly, via XBase.
-
-=head1 DESCRIPTION
-
-Objects of this class are created to deal with memo files, currently
-.dbt. Defines method B<read_header> to parse that header of the file
-and set object's structures, B<write_record> and B<last_record> to
-work properly on this type of file.
-
-There are two separate packages in this module, XBase::Memo::dBaseIII
-and XBase::Memo::dBaseIV. Memo objects are effectively of one of these
-types and they specify B<read_record> and B<write_record> methods.
-
-=head1 VERSION
-
-0.0584
-
-=head1 AUTHOR
-
-(c) Jan Pazdziora, adelton@fi.muni.cz
-
-=head1 SEE ALSO
-
-perl(1), XBase(3)
-
 =cut
 
-
-# ###################################
-# Here starts the XBase::Memo package
-
 package XBase::Memo;
 
 use strict;
 use XBase::Base;
 
-
 use vars qw( $VERSION @ISA );
 @ISA = qw( XBase::Base );
+$VERSION = 0.0597;
 
-
-$VERSION = "0.0584";
-
+# Read header is called from open to fill the object structures
 sub read_header
 	{
-	my $self = shift;
+	my ($self, $dbf_version) = (shift, shift);
 
 	my $header;
 	$self->{'fh'}->read($header, 512) == 512 or do
-		{ Error "Error reading header of $self->{'filename'}\n";
-		return; };
+		{ $self->Error("Error reading header of $self->{'filename'}: $!\n"); return; };
 
 	my ($next_for_append, $block_size, $version);
 	if ($self->{'filename'} =~ /\.fpt$/i)
@@ -67,10 +33,13 @@ sub read_header
 	else
 		{
 		($next_for_append, $version, $block_size)
-					= unpack 'V@16C@20v', $header;
-		if ($version == 3)
+					= unpack 'V @16C @20v', $header;
+		if ($version == 3 or
+			($version == 0 and defined $dbf_version and
+				(($dbf_version & 7) == 3)))
 			{
 			$block_size = 512;
+			$version = 3;
 			bless $self, 'XBase::Memo::dBaseIII';
 			}
 		else
@@ -91,11 +60,11 @@ sub read_header
 sub write_record
 	{
 	my ($self, $num) = (shift, shift);
-	my $length = length(join "", @_);
+	my $length = length(join '', @_);
 	my $record_len = $self->{'record_len'};
 	my $num_of_blocks = int (($length + $record_len - 1) / $record_len);
 	$self->SUPER::write_record($num, @_);
-	if ($num > $self->last_record())
+	if ($num < 0 or $num > $self->last_record())
 		{
 		$self->SUPER::write_to(0, pack "V", $num + $num_of_blocks);
 		$self->{'next_for_append'} = $num + $num_of_blocks;
@@ -114,7 +83,7 @@ sub create
 	$version = 3 unless defined $version;
 	$version = 0 if $version == 4;
 	my $header = 
-	$self->write_to(0, pack "VVa8Ca495", 1, 0, '', $version) or return;
+	$self->write_to(0, pack 'VVa8Ca495', 1, 0, '', $version) or return;
 	$self->close();
 	return $self;
 	}
@@ -136,9 +105,8 @@ sub read_record
 	my $last = $self->last_record();
 	while ($num <= $last)
 		{
-		my $buffer = $self->SUPER::read_record($num, -1);
-		if (not defined $buffer) { return; }
-		my $index = index($buffer, "\x1a\x1a");
+		my $buffer = $self->SUPER::read_record($num, -1) or return;
+		my $index = index($buffer, "\x1a");
 		if ($index >= 0)
 			{ return $result . substr($buffer, 0, $index); }
 		$result .= $buffer;
@@ -151,7 +119,7 @@ sub write_record
 	my ($self, $num) = (shift, shift);
 	my $type = shift;
 	my $data = join "", @_, "\x1a\x1a";
-	if ($num < $self->last_record() and $num != -1)
+	if ($num > 0 and $num < $self->last_record())
 		{
 		my $buffer = $self->read_record($num);
 		if (defined $buffer)
@@ -167,9 +135,7 @@ sub write_record
 			}
 		}
 	else
-		{
-		$num = $self->last_record() + 1;
-		}
+		{ $num = $self->last_record() + 1; }
 	$self->SUPER::write_record($num, $data);
 	$num;
 	}
@@ -187,18 +153,18 @@ sub read_record
 	{
 	my ($self, $num) = @_;
 	my $result = '';
-	my $last = $self->last_record();
+	my $last = $self->last_record;
 
 	my $buffer = $self->SUPER::read_record($num, -1);
 	if (not defined $buffer) { return; }
-	my ($unused_id, $length) = unpack "VV", $buffer;
+	my ($unused_id, $length) = unpack 'VV', $buffer;
 	my $block_size = $self->{'record_len'};
 	if ($length < $block_size - 8)
 		{ return substr $buffer, 8, $length; }
 	my $rest_length = $length - ($block_size - 8);
 	my $rest_data = $self->SUPER::read_record($num + 1, $rest_length);
 	if (not defined $rest_data) { return; }
-	return substr($buffer, 8, $block_size - 8) . $rest_data;
+	return substr($buffer, 8) . $rest_data;
 	}
 
 sub write_record
@@ -252,3 +218,32 @@ use vars qw( @ISA );
 
 1;
 
+__END__
+
+=head1 SYNOPSIS
+
+Used indirectly, via XBase. Users should check its man page.
+
+=head1 DESCRIPTION
+
+Objects of this class are created to deal with memo files, currently
+.dbt and .fpt. Defines method B<read_header> to parse that header of
+the file and set object's structures, B<write_record> and
+B<last_record> to work properly on these types of file.
+
+There are three separate subpackages in XBase::Memo, dBaseIII, dBaseIV
+and Fox. Memo objects are effectively of one of these types and they
+specify their specific B<read_record> and B<write_record> methods.
+
+=head1 VERSION
+
+0.0597
+
+=head1 AUTHOR
+
+(c) Jan Pazdziora, adelton@fi.muni.cz
+
+=head1 SEE ALSO
+
+perl(1), XBase(3)
+
